#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#define IN 0
#define OUT 1

int * pipepair()
{
	int * pipes = calloc(2, sizeof(int));
	if(pipe(pipes) != 0)
	{
		free(pipes);
		pipes = NULL;
	}
	return pipes;
}

//Message Handler type
typedef void (*message_handler)(void*);

//Message handlers
void default_handler(void* unused)
{
	fprintf(stderr, "Unrecognised message type received\n");
}

void add(int * numbers)
{
	printf("%d + %d = %d\n", numbers[0], numbers[1], numbers[0] + numbers[1]);
}

void subtract(int * numbers)
{
	printf("%d - %d = %d\n", numbers[0], numbers[1], numbers[0] - numbers[1]);
}

void message_loop(message_handler* handlers, int pipe)
{
	char type;
	unsigned int length;
	unsigned int buffersize = 1024;
	char * message = malloc(1024);
	struct iovec header[2] = {{&type, sizeof(type)},{&length, sizeof(length)}};
	while(1)
	{
		readv(pipe, header, 2);
		if(length > buffersize)
		{
			buffersize = length;
			message = realloc(message, buffersize);
		}
		if(type == 0)
		{
			return;
		}
		read(pipe, message, length);
		handlers[type](message);
	}
}



void send_message(int pipe, char type, char * body, unsigned int length)
{
	struct iovec message[3] = {{&type, 1}, {&length, sizeof(length)}, {body, length}};
	writev(pipe, message, 3);
}

message_handler * new_handlers(void)
{
	message_handler * handlers = calloc(255, sizeof(message_handler));
	for(unsigned int i=0 ; i<255 ; i++)
	{
		handlers[i] = default_handler;
	}
	return handlers;
}

int main(void)
{
	int * pipes = pipepair();
	if(pipes == NULL)
	{
		return -1;
	}
	pid_t pid = fork();
	if(pid == 0)
	{
		message_handler * handlers = new_handlers();
		handlers[1] = (message_handler)add;
		handlers[2] = (message_handler)subtract;
		message_loop(handlers, pipes[IN]);
	}
	else
	{
		int body[2] = {8,6};
		send_message(pipes[OUT], 1, (void*)body, 2*sizeof(int));
		send_message(pipes[OUT], 2, (void*)body, 2*sizeof(int));
		send_message(pipes[OUT], 0, NULL, 0);
	}
	return 0;
}
